package wacc

import cats.syntax.all._
import scala.collection.mutable

object typeChecker {
  import wacc.ast._
  import wacc.types._

  case class TypeCheckerCtx(
      globalNames: Map[Ident, SemType],
      errors: mutable.Builder[Error, List[Error]]
  ) {
    def typeOf(ident: Ident): SemType = globalNames.withDefault { case Ident(_, -1) => ? }(ident)

    def error(err: Error): SemType =
      errors += err
      ?
  }

  enum Constraint {
    case Unconstrained
    case Is(ty: SemType, msg: String)
    case IsSymmetricCompatible(ty: SemType, msg: String)
    case IsUnweakanable(ty: SemType, msg: String)
    case IsVar(msg: String)
    case IsEither(ty1: SemType, ty2: SemType, msg: String)
    case Never(msg: String)
  }

  extension (ty: SemType) {
    def satisfies(constraint: Constraint, pos: Position)(using ctx: TypeCheckerCtx): SemType =
      (ty, constraint) match {
        case (KnownType.Array(KnownType.Char), Constraint.Is(KnownType.String, _)) =>
          KnownType.String
        case (
              KnownType.String,
              Constraint.IsSymmetricCompatible(KnownType.Array(KnownType.Char), _)
            ) =>
          KnownType.String
        case (ty, Constraint.IsSymmetricCompatible(ty2, msg)) =>
          ty.satisfies(Constraint.Is(ty2, msg), pos)
        case (ty, Constraint.Is(ty2, msg)) => ty.satisfies(Constraint.IsUnweakanable(ty2, msg), pos)
        case (ty, Constraint.Unconstrained) => ty
        case (KnownType.Func(_, _), Constraint.IsVar(msg)) =>
          ctx.error(Error.SemanticError(pos, msg))
        case (ty, Constraint.IsVar(msg)) => ty
        case (ty, Constraint.Never(msg)) =>
          ctx.error(Error.SemanticError(pos, msg))
        case (ty, Constraint.IsEither(ty1, ty2, msg)) =>
          (ty moreSpecific ty1).orElse(ty moreSpecific ty2).getOrElse {
            ctx.error(Error.TypeMismatch(pos, ty1, ty, msg))
          }
        case (ty, Constraint.IsUnweakanable(ty2, msg)) =>
          (ty moreSpecific ty2).getOrElse {
            ctx.error(Error.TypeMismatch(pos, ty2, ty, msg))
          }
      }

    infix def moreSpecific(ty2: SemType): Option[SemType] =
      (ty, ty2) match {
        case (ty, ?)                  => Some(ty)
        case (?, ty)                  => Some(ty)
        case (ty1, ty2) if ty1 == ty2 => Some(ty1)
        case (KnownType.Array(inn1), KnownType.Array(inn2)) =>
          (inn1 moreSpecific inn2).map(KnownType.Array(_))
        case (KnownType.Pair(fst1, snd1), KnownType.Pair(fst2, snd2)) =>
          (fst1 moreSpecific fst2, snd1 moreSpecific snd2).mapN(KnownType.Pair(_, _))
        case _ => None
      }
  }

  def check(prog: Program)(using
      ctx: TypeCheckerCtx
  ): Unit = {
    prog.funcs.foreach { case FuncDecl(_, name, _, stmts) =>
      ctx.typeOf(name) match {
        case KnownType.Func(retType, _) =>
          stmts.toList.foreach(
            checkStmt(_, Constraint.Is(retType, s"function ${name.v} must return $retType"))
          )
        case _ => ctx.error(Error.InternalError(name.pos, "function declaration with non-function"))
      }
    }
    prog.main.toList.foreach(checkStmt(_, Constraint.Never("main function must not return")))
  }

  private def checkStmt(stmt: Stmt, returnConstraint: Constraint)(using
      ctx: TypeCheckerCtx
  ): Unit = stmt match {
    case VarDecl(_, name, value) =>
      val expectedTy = ctx.typeOf(name)
      checkValue(
        value,
        Constraint.Is(
          expectedTy,
          s"variable ${name.v} must be assigned a value of type $expectedTy"
        )
      )
    case Assign(lhs, rhs) =>
      val lhsTy = checkValue(lhs, Constraint.Unconstrained)
      checkValue(rhs, Constraint.Is(lhsTy, s"assignment must have type $lhsTy")) match {
        case ? =>
          ctx.error(
            Error.SemanticError(lhs.pos, "assignment with both sides of unknown type is illegal")
          )
        case _ => ()
      }
    case Read(lhs) =>
      val lhsTy = checkValue(lhs, Constraint.Unconstrained)
      lhsTy match {
        case ? =>
          ctx.error(
            Error.SemanticError(lhs.pos, "cannot read into a destination with an unknown type")
          )
        case _ =>
          lhsTy.satisfies(
            Constraint.IsEither(
              KnownType.Int,
              KnownType.Char,
              "read must be applied to an int or char"
            ),
            lhs.pos
          )
      }
    case Free(lhs) =>
      checkValue(
        lhs,
        Constraint.IsEither(
          KnownType.Array(?),
          KnownType.Pair(?, ?),
          "free must be an array or pair"
        )
      )
    case Return(expr) =>
      checkValue(expr, returnConstraint)
    case Exit(expr) =>
      checkValue(expr, Constraint.Is(KnownType.Int, "exit value must be int"))
    case Print(expr, _) =>
      // This constraint should never fail, the scope-checker should have caught it already
      checkValue(expr, Constraint.IsVar("print value must be a variable"))
    case If(cond, thenStmt, elseStmt) =>
      checkValue(cond, Constraint.Is(KnownType.Bool, "if condition must be a bool"))
      thenStmt.toList.foreach(checkStmt(_, returnConstraint))
      elseStmt.toList.foreach(checkStmt(_, returnConstraint))
    case While(cond, body) =>
      checkValue(cond, Constraint.Is(KnownType.Bool, "while condition must be a bool"))
      body.toList.foreach(checkStmt(_, returnConstraint))
    case Block(body) =>
      body.toList.foreach(checkStmt(_, returnConstraint))
    case Skip() => ()
  }

  private def checkValue(value: LValue | RValue | Expr, constraint: Constraint)(using
      ctx: TypeCheckerCtx
  ): SemType = value match {
    case l @ IntLiter(_)  => KnownType.Int.satisfies(constraint, l.pos)
    case l @ BoolLiter(_) => KnownType.Bool.satisfies(constraint, l.pos)
    case l @ CharLiter(_) => KnownType.Char.satisfies(constraint, l.pos)
    case l @ StrLiter(_)  => KnownType.String.satisfies(constraint, l.pos)
    case l @ PairLiter()  => KnownType.Pair(?, ?).satisfies(constraint, l.pos)
    case id: Ident =>
      ctx.typeOf(id).satisfies(constraint, id.pos)
    case ArrayElem(id, indices) =>
      val arrayTy = ctx.typeOf(id)
      val elemTy = indices.toList.foldRight(arrayTy) { (elem, acc) =>
        checkValue(elem, Constraint.Is(KnownType.Int, "array index must be an int"))
        acc match {
          case KnownType.Array(innerTy) => innerTy
          case _ =>
            ctx.error(
              Error.TypeMismatch(elem.pos, KnownType.Array(?), acc, "cannot index into a non-array")
            )
        }
      }
      elemTy.satisfies(constraint, id.pos)
    case Parens(expr) => checkValue(expr, constraint)
    case l @ ArrayLiter(elems) =>
      KnownType
        .Array(elems.foldRight[SemType](?) { case (elem, acc) =>
          checkValue(
            elem,
            Constraint.IsSymmetricCompatible(acc, "array elements must have the same type")
          )
        })
        .satisfies(constraint, l.pos)
    case l @ NewPair(fst, snd) =>
      KnownType
        .Pair(
          checkValue(fst, Constraint.Unconstrained),
          checkValue(snd, Constraint.Unconstrained)
        )
        .satisfies(constraint, l.pos)
    case Call(id, args) =>
      val funcTy = ctx.typeOf(id)
      funcTy match {
        case KnownType.Func(retTy, paramTys) =>
          if (args.length != paramTys.length) {
            ctx.error(Error.FunctionParamsMismatch(id.pos, paramTys.length, args.length))
          }
          args.zip(paramTys).foreach { case (arg, paramTy) =>
            checkValue(arg, Constraint.Is(paramTy, s"argument type mismatch in function ${id.v}"))
          }
          retTy.satisfies(constraint, id.pos)
        // Should never happen, the scope-checker should have caught this already
        // ctx error had it not
        case _ => ctx.error(Error.InternalError(id.pos, "function call to non-function"))
      }
    case Fst(elem) =>
      checkValue(
        elem,
        Constraint.Is(KnownType.Pair(?, ?), "fst must be applied to a pair")
      ) match {
        case what @ KnownType.Pair(left, _) =>
          left.satisfies(constraint, elem.pos)
        case ? => ?.satisfies(constraint, elem.pos)
        case _ => ctx.error(Error.InternalError(elem.pos, "fst must be applied to a pair"))
      } // satisfies constraint
    case Snd(elem) =>
      checkValue(
        elem,
        Constraint.Is(KnownType.Pair(?, ?), "snd must be applied to a pair")
      ) match {
        case KnownType.Pair(_, right) => right.satisfies(constraint, elem.pos)
        case ?                        => ?.satisfies(constraint, elem.pos)
        case _ => ctx.error(Error.InternalError(elem.pos, "snd must be applied to a pair"))
      }

    // Unary operators
    case Negate(x) =>
      checkValue(x, Constraint.Is(KnownType.Int, "negation must be applied to an int"))
      KnownType.Int.satisfies(constraint, x.pos)
    case Not(x) =>
      checkValue(x, Constraint.Is(KnownType.Bool, "logical not must be applied to a bool"))
      KnownType.Bool.satisfies(constraint, x.pos)
    case Len(x) =>
      checkValue(x, Constraint.Is(KnownType.Array(?), "len must be applied to an array"))
      KnownType.Int.satisfies(constraint, x.pos)
    case Ord(x) =>
      checkValue(x, Constraint.Is(KnownType.Char, "ord must be applied to a char"))
      KnownType.Int.satisfies(constraint, x.pos)
    case Chr(x) =>
      checkValue(x, Constraint.Is(KnownType.Int, "chr must be applied to an int"))
      KnownType.Char.satisfies(constraint, x.pos)

    // Binary operators
    case op: (Add | Sub | Mul | Div | Mod) =>
      val operand = Constraint.Is(KnownType.Int, "binary operator must be applied to an int")
      checkValue(op.x, operand)
      checkValue(op.y, operand)
      KnownType.Int.satisfies(constraint, op.pos)
    case op: (Eq | Neq) =>
      val xTy = checkValue(op.x, Constraint.Unconstrained)
      checkValue(op.y, Constraint.Is(xTy, "equality must be applied to values of the same type"))
      KnownType.Bool.satisfies(constraint, op.pos)
    case op: (Less | LessEq | Greater | GreaterEq) =>
      val xTy = checkValue(
        op.x,
        Constraint.IsEither(
          KnownType.Int,
          KnownType.Char,
          "comparison must be applied to an int or char"
        )
      )
      checkValue(op.y, Constraint.Is(xTy, "comparison must be applied to values of the same type"))
      KnownType.Bool.satisfies(constraint, op.pos)
    case op: (And | Or) =>
      val operand = Constraint.Is(KnownType.Bool, "logical operator must be applied to a bool")
      checkValue(op.x, operand)
      checkValue(op.y, operand)
      KnownType.Bool.satisfies(constraint, op.pos)
  }
}
